import argparse
import mouette as M
from mouette.processing import SurfaceFrameField, PrincipalDirections

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("input_mesh", type=str, help="path to the input mesh")
    parser.add_argument("-outp", "--outp", default="output/output.geogram_ascii")
    parser.add_argument("-v", "--verbose", action="store_true")
    parser.add_argument("-elem", "--elements", type=str, choices=["vertices", "faces"])
    parser.add_argument("-order", "--order", type=int, default=4)
    parser.add_argument("-feat", "--features", action="store_true")
    parser.add_argument("-curv", "--curvature", action="store_true")
    parser.add_argument("-cadff", "--cadff", action="store_true")
    parser.add_argument("-singus", "--singular-indices", nargs="+", help="list of alternating elem_id singu_id")
    parser.add_argument("-n", "--n-smooth", default=0, type=int)
    parser.add_argument("-alpha", "--alpha", default=None, type=float)
    args = parser.parse_args()

    OUTPUT_FILE = args.outp
    PREV = OUTPUT_FILE.split(".")
    OUTPUT_FF  = PREV[0] + "_FF." + PREV[1]

    mesh = M.mesh.load(args.input_mesh)
    
    ## Handle provided singularities
    singus = None
    if args.singular_indices is not None:
        assert len(args.singular_indices)%2 == 0
        singus = mesh.vertices.create_attribute("singuls", float) if args.elem == "faces" else mesh.faces.create_attribute("singuls", float)
        for i in range(len(args.singular_indices)//2):
            singus[args.singular_indices[2*i]] = args.singular_indices[2*i+1]

    if args.curvature:
        ff = PrincipalDirections(mesh, args.elements, args.features, args.verbose, n_smooth=args.n_smooth, smooth_attach_weight=args.alpha)
    else:
        ff = SurfaceFrameField(mesh, args.elements, args.order, args.features, verbose=args.verbose, n_smooth=args.n_smooth, smooth_attach_weight=args.alpha, cad_correction=args.cadff, singularity_indices=singus)

    ff.run()
    ff.flag_singularities()
    M.mesh.save(mesh, OUTPUT_FILE)
    M.mesh.save(ff.export_as_mesh(), OUTPUT_FF)



from ..mesh_data import RawMeshData
from ...geometry import Vec
from ..mesh import PointCloud

def import_xyz(filepath : str) -> RawMeshData:
    """Imports a point cloud from a .xyz file on the disk

    Parameters:
        filepath (str): path to the .xyz file

    Returns:
        RawMeshData: parsed file
    """
    obj = RawMeshData()
    with open(filepath, 'r') as f:
        for v in f.readlines():
            v = [float(x) for x in v.strip().split()[:3]]
            obj.vertices.append(v)
    return obj

def export_xyz(mesh : PointCloud, filepath : str) -> None:
    """Exports a Point Cloud to a .xyz file

    Parameters:
        mesh (PointCloud): the object to be exported
        filepath (str): path to the .xyz file
    """
    with open(filepath, 'w') as f:
        for v in mesh.vertices:
            f.write("{} {} {}\n".format(v[0], v[1], v[2]))


from ...mesh.datatypes import *
from ...mesh.mesh import copy
from ...mesh.mesh_attributes import Attribute, ArrayAttribute
from ...geometry import Vec
from ... import operators
from ..border import extract_border_cycle
from ..worker import Worker
from ...attributes.glob import euler_characteristic
from ...utils.argument_check import check_argument

import numpy as np
from enum import Enum
import cmath
from math import pi
from scipy.sparse import linalg


class TutteEmbedding(Worker):
    """
    Tutte's embedding parametrization method for a disk inside a fixed boundary.
    The parametrization is locally injective (Floater, 1997) provided the boundary is convex.

    References:
        How to draw a graph, Tutte, 1963.
    """

    class BoundaryMode(Enum):
        """
        Enum that represents the shape of the boundary curve.
        """
        CIRCLE = 0
        SQUARE = 1

        @staticmethod
        def from_string(s :str):
            if "circle" in s.lower():
                return TutteEmbedding.BoundaryMode.CIRCLE
            if "square" in s.lower():
                return TutteEmbedding.BoundaryMode.SQUARE
            raise Exception(f"'{s}' does correspond to any boundary mode. Choices are : 'circle', 'square'")

    def __init__(self, mesh : SurfaceMesh, verbose : bool=True, boundary_mode : str = "circle", save_on_mesh : bool = True):
        """
        Initializer of the Tutte's embedding tool.

        Args:
            mesh (SurfaceMesh): the mesh to embed. Should be a surface with disk topology.
            verbose (bool, optional): verbose mode. Defaults to True.
            boundary_mode (str, optional): Shape of the boundary. Possible choices are ["square", "circle"]. Defaults to "circle".
            save_on_mesh (bool, optional): whether to write uv coordinates as an attribute on the mesh. Defaults to True.
        

        Raises:
            InvalidArgumentValueError : if 'boundary_mode' is not "square" or "circle".
        """
        check_argument("boundary_mode", boundary_mode, str, ["square", "circle"])
        super().__init__(self, verbose=verbose)
        self.mesh : SurfaceMesh = mesh
        self._bnd_mode : TutteEmbedding.BoundaryMode = TutteEmbedding.BoundaryMode.from_string(boundary_mode)
        self._save_on_mesh : bool = save_on_mesh
        self._flat_mesh : SurfaceMesh = None
        self.uvs : Attribute = None # attribute on vertices

    def run(self) :
        if euler_characteristic(self.mesh)!=1:
            raise Exception("Mesh is not a topological disk. Cannot run parametrization.")

        Ybnd = self._initialize_boundary(self._bnd_mode)

        lap = operators.laplacian(self.mesh, cotan=False)
        freeInds = self.mesh.interior_vertices
        bndInds, _ = extract_border_cycle(self.mesh)

        LI = lap[freeInds, :][:, freeInds]
        LB = lap[freeInds, :][:, bndInds]

        U = linalg.spsolve(LI, -LB.dot(Ybnd[:,0]))
        V = linalg.spsolve(LI, -LB.dot(Ybnd[:,1]))

        if self._save_on_mesh:
            self.uvs = self.mesh.vertices.create_attribute("uv_coords", float, 2, dense=True)
        else:
            self.uvs = ArrayAttribute(float, len(self.mesh.vertices), 2)
        for i,v in enumerate(freeInds):
            self.uvs[v] = Vec(U[i], V[i])
        for i,v in enumerate(bndInds):
            self.uvs[v] = Vec(Ybnd[i,:])

    def _initialize_boundary(self, boundary_mode):
        n = len(self.mesh.boundary_vertices)
        Ybnd = np.zeros((n,2))
        if boundary_mode == TutteEmbedding.BoundaryMode.CIRCLE:
            for i in range(n):
                rt = cmath.rect(1., 2*pi*i/n)
                Ybnd[i,0] = rt.real
                Ybnd[i,1] = rt.imag
        elif boundary_mode == TutteEmbedding.BoundaryMode.SQUARE:
            corners = [0, n//4, n//2, (3*n)//4]
            Ybnd[corners[0], :] = [0,0]
            Ybnd[corners[1], :] = [1,0]
            Ybnd[corners[2], :] = [1,1]
            Ybnd[corners[3], :] = [0,1]
            for i in range(1, n//4):
                Ybnd[i,:] = Vec(4*i/n, 0)
            for i,v in enumerate(range(n//4+1, n//2)):
                Ybnd[v,:] = Vec(1, 4*i/n)
            for i,v in enumerate(range(n//2+1, (3*n)//4)):
                Ybnd[v,:] = Vec(1-4*i/n, 1)
            for i,v in enumerate(range((3*n)//4+1, n)):
                Ybnd[v,:] = Vec(0, 1-4*i/n)
        return Ybnd

    @property
    def flat_mesh(self):
        """
        A flat representation of the mesh where uv-coordinates are copied to xy.

        Returns:
            SurfaceMesh: the flat mesh
        """
        if self.uvs is None:
            return None
        if self._flat_mesh is None:
            # build the flat mesh : vertex coordinates are uv of original mesh
            self._flat_mesh = copy(self.mesh)
            for v in self.mesh.id_vertices:
                self._flat_mesh.vertices[v] = Vec(self.uvs[v][0], self.uvs[v][1], 0)
        return self._flat_mesh


import mouette as M
import numpy as np

def compare_container(cont1, cont2):
    if len(cont1) != len(cont2): 
        return False
    for i in range(len(cont1)):
        if isinstance(cont1[i], np.ndarray):
            if (cont1[i] != cont2[i]).all():
                return False
    return True

def compare_mesh(mesh1, mesh2):
    if type(mesh1) != type(mesh2) : return False
    comp = compare_container(mesh1.vertices, mesh2.vertices)
    if hasattr(mesh1, "edges"):
        comp = comp and compare_container(mesh1.edges, mesh2.edges)
    if hasattr(mesh1, "faces"):
        comp = comp and compare_container(mesh1.faces, mesh2.faces)
    if hasattr(mesh1, "cells"):
        comp = comp and compare_container(mesh1.cells, mesh2.cells)
    return comp

def build_test_io(mesh, tmp_path, fmt, dim=None):
    filepath = tmp_path / f"test_mesh.{fmt}"
    M.mesh.save(mesh, str(filepath))
    mesh2 = M.mesh.load(str(filepath), dim)
    return compare_mesh(mesh, mesh2)


from ..mesh.datatypes import *
from ..mesh.mesh import _instanciate_raw_mesh_data
from ..mesh.mesh_data import RawMeshData
from ..geometry import Vec
from math import sqrt

def tetrahedron(P1,P2,P3,P4, volume=False) -> SurfaceMesh:
    tet = RawMeshData()
    tet.vertices += [P1,P2,P3,P4]
    tet.faces += [(1,2,3), (0,2,3),(0,1,3),(0,1,2)]
    if volume: tet.cells.append((0,1,2,3))
    return _instanciate_raw_mesh_data(tet)

def hexahedron(P1, P2, P3, P4, P5, P6, P7, P8, colored=False, volume=False) -> SurfaceMesh:
    """Generate an hexahedron in arbitrary configuration given 8 points. Order and connectivity of points is:

       8--------7
      /|       /|
     / |      / |
    5--------6  |
    |  |     |  |
    |  4-----|--3
    | /      | /
    |/       |/
    1--------2

    Parameters:
        P1 to P8 (Vec) : coordinates of height vertices
        colored (bool, optional): if set to true, will add a color attribute on faces. Defaults to False.
        volume (bool, optional): if set to true, will also generate three tetrahedra to fill the volume. Defaults to False.

    Returns:
        SurfaceMesh: a cube
    """
    hexa = RawMeshData()
    hexa.vertices += [P1,P2,P3,P4,P5,P6,P7,P8]

    if volume:
        #hexa.cells += [(0,1,2,3),(0,5,7,4), (2,6,7,5)]
        hexa.cells += [(0,1,2,3,4,5,6,7)]
    else:
        hexa.faces += [
            (0,2,1), (0,3,2),
            (0,1,5), (0,5,4),
            (1,2,6), (1,6,5),
            (2,3,7), (2,7,6),
            (3,0,4), (3,4,7),
            (4,5,6), (4,6,7)
        ]
        if colored:
            col = hexa.faces.create_attribute("color", float, 3)
            RED,GREEN,BLUE = Vec(1.,0.,0), Vec(0.,1.,0.), Vec(0.,0.,1.)
            col[0] = RED
            col[1] = RED
            col[10] = RED
            col[11] = RED

            col[2] = GREEN
            col[3] = GREEN
            col[6] = GREEN
            col[7] = GREEN
            
            col[4] = BLUE
            col[5] = BLUE
            col[8] = BLUE
            col[9] = BLUE
    return _instanciate_raw_mesh_data(hexa)

def octahedron():
    raise NotImplementedError

def icosahedron(center : Vec = Vec(0,0,0), uv=False):
    phi = (1 + sqrt(5)) / 2
    m = RawMeshData()

    m.vertices += [ a+center for a in 
    [
        Vec(-1, phi,0),
        Vec(1, phi, 0),
        Vec(-1, -phi, 0),
        Vec(1, -phi, 0),

        Vec(0, -1, phi),
        Vec(0, 1, phi),
        Vec(0, -1, -phi),
        Vec(0, 1, -phi),

        Vec(phi, 0, -1),
        Vec(phi, 0, 1),
        Vec(-phi, 0, -1),
        Vec(-phi, 0, 1),
    ]]

    m.faces += [
        (0, 11, 5),
        (0, 5, 1),
        (0, 1, 7),
        (0, 7, 10),
        (0, 10, 11),
        (1, 5, 9),
        (5, 11, 4),
        (11, 10, 2),
        (10, 7, 6),
        (7, 1, 8),
        (3, 9, 4),
        (3, 4, 2),
        (3, 2, 6),
        (3, 6, 8),
        (3, 8, 9),
        (4, 9, 5),
        (2, 4, 11),
        (6, 2, 10),
        (8, 6, 7),
        (9, 8, 1)
    ]
    return _instanciate_raw_mesh_data(m, 2)

def dodecahedron() -> SurfaceMesh:
    raise NotImplementedError


import scipy.sparse as sp
import numpy as np
from .. import geometry as geom

def inverse_power_method(A : sp.csc_matrix, B: sp.csc_matrix = None, maxiter:int = 100, tol:float=1e-6) -> np.ndarray:
    """
    Implementation of the inverse power method (or inverse iteration) scheme to find an eigenvector associated with the smallest eigenvalue of A.
    In other words, this function computes x such that : Ax = λBx where λ is the smallest (non-zero) eigenvalue (in magnitude) of A.

    Args:
        A (sp.csc_matrix): the matrix 
        B (sp.csc_matrix, optional): metrics matrix. If not specified, will be the identity matrix. Defaults to None.
        maxiter (int, optional): maximal number of internal iteration. Defaults to 100.
        tol (float, optional): early stopping criterion. Will stop the iteration if |x_{n+1} - x_n| < tol. Defaults to 1e-6.

    Returns:
        np.ndarray: _description_
    """
    n = A.shape[0]
    if B is None : B = sp.eye(n, format="csc")
    solve = sp.linalg.factorized(A)
    x = np.random.random(n)
    it = 0
    stop_criterion = False
    while not stop_criterion:
        newx = solve(B@x)
        newx /= np.sqrt(np.dot(newx, B@newx))
        it += 1
        stop_criterion = (it>=maxiter or geom.distance(newx, x)<tol)
    return x



"""
Mouette example files.
Parametrization using LSCM
"""

import mouette as M
from mouette.processing import parametrization
import sys

if len(sys.argv)<2:
    print("Usage: `python lscm.py <path/to/mesh>")

mesh = M.mesh.load(sys.argv[1])
lscm = parametrization.LSCM(mesh, verbose=True)()

M.mesh.save(mesh, "output.geogram_ascii")
M.mesh.save(lscm.flat_mesh, "output_flat.geogram_ascii")


import numpy as np

from ..geometry import *
from ..attributes import mean_edge_length
from ..mesh.datatypes import *
from ..mesh.mesh_data import RawMeshData
from ..mesh.mesh import merge, _instanciate_raw_mesh_data

def cylinder(P1 : Vec, P2 : Vec, radius: float = 1., N=50, fill_caps=True) -> SurfaceMesh:
    cy = RawMeshData()
    axis = Vec.normalized(P2-P1)
    t = Vec(axis.y, -axis.x, 0.) # tangent vector
    if t.norm()<1e-6:
        t = Vec(0, axis.z, -axis.y) # another tangent vector
    t = Vec.normalized(t)
    for P in (P1,P2):
        for i in range(N):
            Pi = P + radius * rotate_around_axis(t, axis, 2*np.pi*i/N)
            cy.vertices.append(Pi)
    # caps
    if fill_caps:
        cy.vertices += [P1,P2]
        for i in range(N):
            cy.faces.append((i,(i+1)%N,2*N))
            cy.faces.append((i+N,2*N+1, (i+1)%N+N))
    # side of cylinder
    for i in range(N):
        cy.faces.append((i,  N+i, (i+1)%N))
        cy.faces.append((N+i, N+(i+1)%N, (i+1)%N))
    return _instanciate_raw_mesh_data(cy, 2)

@forbidden_mesh_types(PointCloud)
def cylindrify_edges( mesh : PolyLine, radius: float = 5e-2, N=50) -> SurfaceMesh:
    """ Transforms edges of a polyline as cylinder surface

    Parameters:
        mesh (PolyLine): the input mesh
        radius (float, optional): Radius of the output cylinders. Defaults to 5e-2.
        N (int, optional): Number of points inside each circle of the cylinders. Defaults to 50.

    Returns:
        SurfaceMesh
    """
    if len(mesh.edges)==0: 
        return SurfaceMesh()

    L = mean_edge_length(mesh)
    cylinders = []
    for A,B in mesh.edges:
        pA,pB = mesh.vertices[A], mesh.vertices[B]
        cylinders.append(cylinder(pA,pB, L*radius, N, fill_caps=False))
    return merge(cylinders)


from .mesh import load, save, copy, merge, new_point_cloud, new_polyline, new_surface, new_volume
from .datatypes import PointCloud, PolyLine, SurfaceMesh, VolumeMesh
from .mesh_data import RawMeshData


from .mesh_data import RawMeshData
from .datatypes import Mesh, PointCloud, PolyLine, SurfaceMesh, VolumeMesh
from .io.io import read_by_extension, write_by_extension

def new_point_cloud() -> PointCloud: 
    """
    Create a new empty point cloud

    Returns:
        PointCloud
    """
    return PointCloud()
    
def new_polyline() -> PolyLine:
    """
    Creates a new empty poly line

    Returns:
        PolyLine
    """
    return PolyLine()

def new_surface() -> SurfaceMesh:
    """
    Creates a new empty surface mesh

    Returns:
        SurfaceMesh
    """
    return SurfaceMesh()

def new_volume() -> VolumeMesh:
    """
    Creates a new empty volume mesh

    Returns:
        VolumeMesh
    """
    return VolumeMesh()


def _instanciate_raw_mesh_data(mesh_data : RawMeshData, dim : int = None) -> Mesh:
    mesh_data.prepare()
    if dim is None: dim = -1
    dim = max(dim, mesh_data.dimensionality)
    if dim==0: return PointCloud(mesh_data)
    if dim==1: return PolyLine(mesh_data)
    if dim==2: return SurfaceMesh(mesh_data)
    if dim==3: return VolumeMesh(mesh_data)

def load(filename : str, dim : int = None, raw : bool = False) -> Mesh:
    """
    Loads a Mesh object from a file on the disk

    Parameters:
        filename (str):  the path of the file
        dim (int, optional): 
            Override for the dimensionality of the resulting mesh.
            Depending on this value, returns a PointCloud, LineMesh, SurfaceMesh or VolumeMesh. 
            If not specified, computed from the data read in the file.
        raw (bool, optional): Returns the RawMeshData object instead of a fully prepared Mesh object. Defaults to False

    Returns:
        Mesh:  A new mesh object (PointCloud, LineMesh, SurfaceMesh or VolumeMesh)
    """
    data = read_by_extension(filename)
    if raw: return data
    return _instanciate_raw_mesh_data(data, dim)
    

def save(mesh : Mesh, filename: str, ignore_elements:set = None) -> None:
    """ Saves the mesh data into a file

    Parameters:
        filename (str): The output file path

    Raises:
        Exception: Unsupported file extension
    """
    if isinstance(mesh,VolumeMesh) and ".geogram" in filename:
        mesh.connectivity._compute_adjacent_cell()
    raw_mesh = RawMeshData(mesh) # get rid of connectivity and additional attributes depending on dimension
    if ignore_elements is not None:
        if "edges" in ignore_elements: raw_mesh.edges.clear()
        if "faces" in ignore_elements:
            raw_mesh.faces.clear()
            raw_mesh.face_corners.clear()
        if "cells" in ignore_elements: 
            raw_mesh.cells.clear()
            raw_mesh.cell_corners.clear()
            raw_mesh.cell_faces.clear()
    write_by_extension(raw_mesh, filename)

def copy(mesh : Mesh, copy_attributes=False, copy_connectivity=False, copy_half_edges=False) -> Mesh:
    """[summary]

    Parameters:
        mesh (Mesh): [description]
        copy_attributes (bool, optional): [description]. Defaults to False.

    Returns:
        Mesh: a hard copy of the given mesh
    """
    from copy import deepcopy
    copy_mesh = type(mesh)() # instanciate a new mesh of the same type

    if copy_attributes:
        copy_mesh.vertices = deepcopy(mesh.vertices)
        if hasattr(mesh, "edges") : copy_mesh.edges = deepcopy(mesh.edges)
        if hasattr(mesh, "faces") :
            copy_mesh.faces = deepcopy(mesh.faces)
            copy_mesh.face_corners = deepcopy(mesh.face_corners)
        if hasattr(mesh, "cells"):
            copy_mesh.cells = deepcopy(mesh.cells)
            copy_mesh.cell_corners = deepcopy(mesh.cell_corners)
    else:
        # copy only _cont data of each container (connectivity data and not attribute data)
        copy_mesh.vertices._data = deepcopy(mesh.vertices._data)
        if hasattr(mesh, "edges"): copy_mesh.edges._data = deepcopy(mesh.edges._data)
        if hasattr(mesh, "faces") : 
            copy_mesh.faces._data = deepcopy(mesh.faces._data)
            copy_mesh.face_corners._data = deepcopy(mesh.face_corners._data)
        if hasattr(mesh, "cells") : 
            copy_mesh.cells._data = deepcopy(mesh.cells._data)
            copy_mesh.cell_corners._data = deepcopy(mesh.cell_corners._data)
        # _cont of face_corners and cell_corners are always empty
    if copy_connectivity and hasattr(mesh, "connectivity"):
        copy_mesh.connectivity = mesh.connectivity
    if copy_half_edges and hasattr(mesh, "half_edges"):
        copy_mesh.half_edges = mesh.half_edges
    return copy_mesh

def merge(mesh_list : list) -> Mesh:
    """Merges a list of independents meshes as a unique mesh.

    Parameters:
        mesh_list (list): the list of meshes. If empty, this function returns None

    Returns:
        Mesh: the merged mesh. Takes the type of the mesh with the largest dimensionality in the list (ie merging a Polyline with a Volume mesh returns a Volume mesh)
    """
    if len(mesh_list)==0: return None
    merged = RawMeshData()
    vertex_offset = 0
    for to_merge in mesh_list:
        merged.vertices += to_merge.vertices
        if hasattr(to_merge, "edges") : 
            merged.edges += [tuple((vertex_offset+u for u in e)) for e in to_merge.edges]
        if hasattr(to_merge, "faces") : 
            merged.faces += [tuple((vertex_offset+u for u in f)) for f in to_merge.faces]
        if hasattr(to_merge, "cells") :
            merged.cells += [tuple((vertex_offset+u for u in c)) for c in to_merge.cells]
        vertex_offset += len(to_merge.vertices)
    return _instanciate_raw_mesh_data(merged)


from scipy.optimize import fmin_l_bfgs_b
import scipy.sparse as sp
import numpy as np

from ..worker import Worker
from ...mesh.datatypes import *
from ...mesh.mesh import copy
from ...mesh.mesh_attributes import Attribute, ArrayAttribute
from ...geometry import Vec
from ... import geometry as geom
from ...utils.argument_check import check_argument
from ...attributes.glob import euler_characteristic
from ...attributes.misc_corners import cotangent
from ...operators import area_weight_matrix

def regul(x, eps=1e-18):
    return (x*x + eps)**(0.25)

def regul_prime(x, eps=1e-18):
    y = x*x + eps
    return 0.5 * x/ (y**0.75)

class CotanEmbedding(Worker):
    """
    Given a parametrization of a disk where boundary is fixed, we can optimize the difference between unsigned and signed areas of triangles to compute a parametrization that is foldover-free.

    References:
        Embedding a triangular graph within a given boundary, Xu et al. (2011)
    """

    @allowed_mesh_types(SurfaceMesh)
    def __init__(self, mesh : SurfaceMesh, uv_attr: Attribute, on_corner:bool, verbose:bool=True, **kwargs):
        """
        Initializing the CotanEmbedding parametrization tool.

        Args:
            mesh (SurfaceMesh): the supporting mesh.
            uv_attr (Attribute): the attribute on vertices/corners that stores uv-coordinates to optimize.
            on_corner (bool): specifies on which primitive the uv_attr is set (corners or vertices)
            verbose (bool, optional): verbose mode. Defaults to True.
        
        Keyword Args:
            mode (str): optimization mode. Either "bfgs" or "alternate". Defaults to "bfgs".
            tutte_if_convex (bool): if the boundary of the shape is convex and mode=="alternate", whether to simply run Tutte's embedding. Defaults to True
            solver_verbose (bool): solver verbose mode. Defaults to False.
        """
        super().__init__("CotanEmbedding", verbose=verbose)
        self.mesh : SurfaceMesh = mesh
        self.flat_mesh : SurfaceMesh = None
        self.uvs : ArrayAttribute = uv_attr
        self.on_corners : bool = on_corner

        self._mode = kwargs.get("mode", "bfgs")
        check_argument("mode", self._mode, str, ["bfgs", "alternate"])
        self._tutte_if_convex : bool = kwargs.get("tutte_if_convex", True)
        self._solver_verbose : bool = kwargs.get("solver_verbose", False)

    def _energy(self,X):
        E = 0
        grad = np.zeros_like(X)
        nv = len(self.mesh.vertices)
        for (A,B,C) in self.mesh.faces:
            Et = 0
            iuA,ivA,iuB,ivB,iuC,ivC = A,nv+A,B,nv+B,C,nv+C
            uA,vA,uB,vB,uC,vC = (X[_x] for _x in (iuA,ivA,iuB,ivB,iuC,ivC))

            Et = uA * (vB - vC) + uB * (vC - vA) + uC * (vA - vB)
            if not self.mesh.is_vertex_on_border(A):
                grad[iuA] += 2*(vB - vC)
                grad[ivA] += 2*(uC - uB)
            if not self.mesh.is_vertex_on_border(B): 
                grad[iuB] += 2*(vC - vA)
                grad[ivB] += 2*(uA - uC)
            if not self.mesh.is_vertex_on_border(C): 
                grad[iuC] += 2*(vA - vB)
                grad[ivC] += 2*(uB - uA)
            
            l_bc = (uB-uC)**2 + (vB-vC)**2
            l_ac = (uA-uC)**2 + (vA-vC)**2  
            l_ab = (uA-uB)**2 + (vA-vB)**2

            L_ab = - l_ab + l_bc + l_ac 
            L_bc =   l_ab - l_bc + l_ac
            L_ac =   l_ab + l_bc - l_ac
            Eli = 2*l_bc*l_ac + 2*l_ac*l_ab + 2*l_bc*l_ab - l_bc*l_bc - l_ac*l_ac - l_ab*l_ab

            El = regul(Eli)
            Elp = regul_prime(Eli)
            if not self.mesh.is_vertex_on_border(A): 
                grad[iuA] += 4*Elp * ( (uA - uB)*L_ab + (uA - uC)*L_ac )
                grad[ivA] += 4*Elp * ( (vA - vB)*L_ab + (vA - vC)*L_ac )
            if not self.mesh.is_vertex_on_border(B): 
                grad[iuB] += 4*Elp * ( (uB - uA)*L_ab + (uB - uC)*L_bc )
                grad[ivB] += 4*Elp * ( (vB - vA)*L_ab + (vB - vC)*L_bc )
            if not self.mesh.is_vertex_on_border(C): 
                grad[iuC] += 4*Elp * ( (uC - uA)*L_ac + (uC - uB)*L_bc )
                grad[ivC] += 4*Elp * ( (vC - vA)*L_ac + (vC - vB)*L_bc )
            E += (El + 2*Et)
        return E, grad

    def run(self):
        """
        Runs the optimization

        Raises:
            Exception: Fails if the mesh is not a topological disk
            Exception: Fails if the mesh is not triangular
        """
        if euler_characteristic(self.mesh)!=1:
            raise Exception("Mesh is not a topological disk. Cannot run parametrization.")
        if not self.mesh.is_triangular():
            raise Exception("Mesh is not triangular.")
        
        # Initialize variable vector from uvs
        nv = len(self.mesh.vertices)
        var = np.zeros(2*nv)
        if self.on_corners:
            for c,v in enumerate(self.mesh.face_corners):
                var[v] = self.uvs[c][0]
                var[nv+v] = self.uvs[c][1]
        else:
            for v in self.mesh.id_vertices:
                var[v] = self.uvs[v][0]
                var[nv+v] = self.uvs[v][1]

        # Call corresponding optimizer
        if self._mode=="bfgs":
            var = self._optimize_bfgs(var, self._solver_verbose) 
        elif self._mode=="alternate":
            var = self._optimize_alternate(var, self._tutte_if_convex, self._solver_verbose)

        # Write uvs in attribute
        if self.on_corners:
            # self.mesh.vertices.delete_attribute("uv_coords")
            for c,v in enumerate(self.mesh.face_corners):
                self.uvs[c] = Vec(var[v], var[nv+v])
        else:
            for v in self.mesh.id_vertices:
                self.uvs[v] = Vec(var[v], var[nv+v])
        self._build_flat_mesh(var)

    def _optimize_bfgs(self, var, solver_verbose:bool):
        # det_init = self.mesh.faces.create_attribute("det_init", float, dense=True)
        # for iT,(A,B,C) in enumerate(self.mesh.faces):
        #     iuA,ivA,iuB,ivB,iuC,ivC = 2*A,2*A+1,2*B,2*B+1,2*C,2*C+1
        #     uA,vA,uB,vB,uC,vC = (var[_x] for _x in (iuA,ivA,iuB,ivB,iuC,ivC))
        #     det_init[iT] = geom.sign(uA * (vB - vC) + uB * (vC - vA) + uC * (vA - vB))

        self.log("Optimize cotan energy")
        # return var
        var,energy,infos = fmin_l_bfgs_b(
            # self._energy_simple,        # function and gradient to optimize
            self._energy,        # function and gradient to optimize
            var,                 # Initial variables
            maxiter=2e4,         # maximum number of iterations
            pgtol=1e-6,          # stops if gradient norm is < pgtol
            m=11,                # number of inner iterations for hessian approx.
            factr=1E6,           # stops if improvement is < factr*EPS
            maxls=30,            # maximum number of linesearch
            disp=solver_verbose, # verbose level
            maxfun=1e7       # maximum number of function evaluation
        )
        self.log(f"Stopped after {infos['nit']} iterations")
        self.log(f"Final energy : {energy}")
        return var

        # Compute dets
        det_final = self.mesh.faces.create_attribute("det_final", float, dense=True)
        for iT,(A,B,C) in enumerate(self.mesh.faces):
            iuA,ivA,iuB,ivB,iuC,ivC = 2*A,2*A+1,2*B,2*B+1,2*C,2*C+1
            uA,vA,uB,vB,uC,vC = (var[_x] for _x in (iuA,ivA,iuB,ivB,iuC,ivC))
            det_final[iT] = geom.sign(uA * (vB - vC) + uB * (vC - vA) + uC * (vA - vB))
        
    def _optimize_alternate(self, var, tutte, verbose):

        def check_dets(var):
            for (A,B,C) in self.mesh.faces:
                uA,vA,uB,vB,uC,vC = (var[_x] for _x in (A,nv+A,B,nv+B,C,nv+C))
                if (uA * (vC - vB) + uB * (vA - vC) + uC * (vB - vA)) < 0 : 
                    return False
            return True

        nv = len(self.mesh.vertices)
        n_iter = 0
        n_iter_max = 100
        n_coeff_lap = 12*len(self.mesh.faces)
        rows = np.zeros(n_coeff_lap, dtype=np.int32)
        cols = np.zeros(n_coeff_lap, dtype=np.int32)
        coeffs = np.zeros(n_coeff_lap, dtype=np.float64)
        Iinds = np.array(self.mesh.interior_vertices, dtype=np.int32)
        Binds = np.array(self.mesh.boundary_vertices, dtype=np.int32)
        AreaW = area_weight_matrix(self.mesh, inverse=True).tocsc()
        if tutte:
            cot = cotangent(self.mesh, persistent=False)

        while n_iter<n_iter_max and not check_dets(var):
            n_iter += 1
            self.log(f"Iter {n_iter}/{n_iter_max}")
            # build cotan laplacian  
            _c = 0
            for iF,(A,B,C) in enumerate(self.mesh.faces):
                if tutte and n_iter==1:
                    cot_a,cot_b,cot_c = (cot[3*iF+_i]/2 for _i in range(3))
                else:
                    uA,vA,uB,vB,uC,vC = (var[_x] for _x in (A,nv+A,B,nv+B,C,nv+C))
                    l_bc = (uB-uC)**2 + (vB-vC)**2
                    l_ac = (uA-uC)**2 + (vA-vC)**2  
                    l_ab = (uA-uB)**2 + (vA-vB)**2
                    Eli = 2*l_bc*l_ac + 2*l_ac*l_ab + 2*l_bc*l_ab - l_bc*l_bc - l_ac*l_ac - l_ab*l_ab
                    # El = np.sqrt(Eli)
                    El =regul(Eli, 1e-18)
                    cot_a = (l_ac+l_ab-l_bc)/El/2
                    cot_b = (l_bc+l_ab-l_ac)/El/2
                    cot_c = (l_bc+l_ac-l_ab)/El/2
                for (i, j, v) in [(A, B, cot_c/2), (B, C, cot_a/2), (C, A, cot_b/2)]:
                    rows[_c], cols[_c], coeffs[_c], _c = i, i, v, _c+1
                    rows[_c], cols[_c], coeffs[_c], _c = j, j, v, _c+1
                    rows[_c], cols[_c], coeffs[_c], _c = i, j, -v, _c+1
                    rows[_c], cols[_c], coeffs[_c], _c = j, i, -v, _c+1 
            L = AreaW * sp.csc_matrix((coeffs,(rows,cols)), dtype= np.float64)
            LI = L[Iinds,:][:,Iinds]
            LB = L[Iinds,:][:,Binds] 
            # solve linear system
            var[Iinds] = sp.linalg.spsolve(LI, -LB.dot(var[Binds])) # first system for u
            var[Iinds+nv] = sp.linalg.spsolve(LI, -LB.dot(var[Binds+nv])) # second system for v
        return var
        
    def _build_flat_mesh(self, var):
        # build the flat mesh : vertex coordinates are uv of original mesh
        self.flat_mesh = copy(self.mesh)
        nv = len(self.mesh.vertices)
        for v in self.mesh.id_vertices:
            self.flat_mesh.vertices[v] = Vec(var[v], var[nv+v], 0.)


import mouette as M
from data import *

@pytest.mark.parametrize("m", polylines + surfaces + volumes)
def test_edge_spanning_tree(m):
    _ = M.processing.trees.EdgeSpanningTree(m)()
    assert True


@pytest.mark.parametrize("m", [surf_circle()])
def test_edge_spanning_tree_avoid_edges(m):
    _ = M.processing.trees.EdgeSpanningTree(m, avoid_boundary=True, avoid_edges={0})()
    assert True

@pytest.mark.parametrize("m", [surf_circle(), surf_half_sphere()])
def test_edge_spanning_tree_traverse(m):
    tree = M.processing.trees.EdgeSpanningTree(m, starting_vertex=0, avoid_boundary=True)
    for _ in tree.traverse(): continue
    assert True

@pytest.mark.parametrize("m", polylines + surfaces + volumes)
def test_edge_minimal_spanning_tree(m):
    _ = M.processing.trees.EdgeMinimalSpanningTree(m, avoid_boundary=True)()
    assert True

@pytest.mark.parametrize("m", polylines + surfaces + volumes)
def test_edge_spanning_forest(m):
    _ = M.processing.trees.EdgeSpanningForest(m)()
    assert True

@pytest.mark.parametrize("m", [surf_circle(), surf_half_sphere()])
def test_edge_spanning_forest_traverse(m):
    forest = M.processing.trees.EdgeSpanningForest(m)()
    for _ in forest.traverse(): continue
    assert True

@pytest.mark.parametrize("m", surfaces)
def test_face_spanning_tree(m):
    tree = M.processing.trees.FaceSpanningTree(m)()
    assert True

@pytest.mark.parametrize("m", [surf_circle()])
def test_face_spanning_tree_avoid_edges(m):
    tree = M.processing.trees.FaceSpanningTree(m,0,{10,12,42})()
    for _ in tree.traverse(): continue
    assert True

@pytest.mark.parametrize("m", surfaces)
def test_face_spanning_forest(m):
    _ = M.processing.trees.FaceSpanningForest(m)()
    assert True

@pytest.mark.parametrize("m", [vol_cuboid()])
def test_cell_spanning_tree(m):
    tree = M.processing.trees.CellSpanningTree(m)()
    assert True
    for _ in tree.traverse(): continue
    assert True

@pytest.mark.parametrize("m", [vol_cuboid()])
def test_cell_spanning_forest(m):
    tree = M.processing.trees.CellSpanningForest(m)()
    assert True
    for _ in tree.traverse(): continue
    assert True


import mouette as M
import pytest
import os
import numpy as np

### Polylines ###

def pl_one_edge():
    m = M.mesh.new_polyline()
    m.vertices += [M.Vec(0.,0.,0.), M.Vec(1.,0.,0.)]
    m.edges.append((0,1))
    return m

def pl_sample1():
    m = M.mesh.new_polyline()
    m.vertices += [
        M.Vec(0.,0.,0.),
        M.Vec(1.,0.,1.),
        M.Vec(2.,1.,0.)
    ]
    m.edges += [(0,1), (1,2)]
    return m

def pl_sample2():
    m = M.mesh.new_polyline()
    m.vertices += [
        M.Vec.random(3) for _ in range(6)
    ]
    m.edges += [(0,1), (1,5), (3,5), (2,3), (0,2), (0,3), (3,4), (2,4)]
    return m

def pl_two_components():
    m = M.mesh.new_polyline()
    m.vertices += [
        M.Vec.random(3) for _ in range(10)
    ]
    m.edges += [(0,1), (1,4), (3,5), (2,3), (0,2), (0,3), (2,4), (6,7), (7,8), (7,9), (6,9)]

    return m

polylines = [
    pl_one_edge(),
    pl_sample1(),
    pl_sample2(),
    pl_two_components()
]


from mouette.processing import parametrization as PARAM
from utils import *
from data import *

@pytest.mark.parametrize("m", [surf_circle(), surf_spline(), surf_half_sphere()])
def test_LSCM(m, tmp_path):
    lscm = PARAM.LSCM(m, verbose=False, eigen=True, save_on_corners=True, solver_verbose=False)
    lscm.run()

@pytest.mark.parametrize("m", [surf_circle(), surf_spline(), surf_half_sphere()])
def test_LSCM_no_eigen(m, tmp_path):
    lscm = PARAM.LSCM(m, verbose=False, eigen=False, save_on_corners=True, solver_verbose=False)
    lscm.run()

@pytest.mark.parametrize("m", [surf_half_sphere()])
def test_export_LSCM(m, tmp_path):
    _ = PARAM.LSCM(m, verbose=False, eigen=False, save_on_corners=True, solver_verbose=False)()
    build_test_io(m, tmp_path, "obj", 2)
    build_test_io(m, tmp_path, "geogram_ascii", 2)


import mouette as M
from data.polylines import *

def test_rings():
    r = M.procedural.ring(10, 0.1, False, 1)
    assert len(r.vertices)==11
    assert len(r.faces) == 10
    r = M.procedural.ring(10, 0.3, True, 2)
    assert len(r.vertices) == 22
    assert len(r.faces) == 20

def test_flat_ring():
    r = M.procedural.flat_ring(10, 0.2, 1)
    assert len(r.vertices) == 12
    assert len(r.faces) == 10

def test_cylinder():
    cy1 = M.procedural.cylinder(M.Vec.random(3), M.Vec.random(3), N=10)
    assert len(cy1.vertices) == 2*10 + 2
    cy2 = M.procedural.cylinder(M.Vec.random(3), M.Vec.random(3), N=20, fill_caps=False)
    assert len(cy2.vertices) == 2*20

@pytest.mark.parametrize("p", polylines)
def test_cylindrify(p):
    m = M.procedural.cylindrify_edges(p, N=10)
    assert isinstance(m, M.mesh.SurfaceMesh)

def test_aacube():
    m = M.procedural.axis_aligned_cube()
    assert len(m.vertices) == 8
    m = M.procedural.axis_aligned_cube(colored=True)
    assert len(m.vertices) == 8

def test_hexa():
    P1, P2, P3, P4 = M.Vec.random(3), M.Vec.random(3), M.Vec.random(3), M.Vec.random(3)
    m = M.procedural.hexahedron_4pts(P1,P2,P3,P4)
    assert len(m.vertices) == 8
    m = M.procedural.hexahedron_4pts(P1,P2,P3,P4, colored=True, volume=True)
    assert len(m.vertices) == 8




from mouette import framefield as ff
from data import *

@pytest.mark.parametrize("m", [vol_cube()])
def test_volume_framefield_vertices(m):
    field = ff.VolumeFrameField(m, "vertices")
    field.run()
    field.flag_singularities()
    assert m.faces.has_attribute("singuls")

@pytest.mark.parametrize("m", [vol_cube()])
def test_volume_framefield_cells(m):
    field = ff.VolumeFrameField(m, "cells")
    field.run()
    field.flag_singularities()
    assert m.vertices.has_attribute("singuls")
    assert m.edges.has_attribute("singuls")



from .vector import Vec
import numpy as np
import math
from typing import Union

def sign0(x : float):
    """sign of x
        1 if x >= 0
        -1 if x < 0

    Parameters:
        x (float): input number

    Returns:
        [float]: the sign of x
    """
    if x>=0: return 1
    return -1

def sign(x : float):
    """sign of x
        1 if x > 0
        -1 if x < 0
        0 if x = 0

    Parameters:
        x (float): input number

    Returns:
        [float]: x/|x|
    """
    if x>0: return 1
    elif x<0 : return -1
    return 0

def norm(x : np.ndarray):
    return np.sqrt(np.dot(x.flatten(), x.flatten()))

def dot(A,B):
    return np.dot(A,B)

def distance(a : np.ndarray, b : np.ndarray):
    return norm(b-a)

def cross(v1,v2):
    return Vec(
        v1[1]*v2[2] - v1[2] * v2[1],
        v2[0]*v1[2] - v2[2] * v1[0],
        v1[0]*v2[1] - v1[1] * v2[0]
    )

def cotan(A:Vec, B:Vec, C:Vec):
    """ 
    A,B,C three points in R^3
    returns the cotangent of the angle from BA to BC
    """
    A, B, C = Vec(A), Vec(B), Vec(C)
    BA = Vec.normalized(A-B)
    BC = Vec.normalized(C-B)
    cosine = np.dot(BA,BC)
    sine = norm(cross(BA,BC))
    return cosine/sine

def angle_3pts(A:Vec, B:Vec, C:Vec) -> float:
    """Angle ABC between three points

    Parameters:
        A (Vec): first point
        B (Vec): central point
        C (Vec): second point

    Returns:
        float: the angle
    """
    BA = Vec(A)-Vec(B)
    BC = Vec(C)-Vec(B)
    s = cross(BA,BC).norm()
    c = dot(BA,BC)
    return math.atan2(s, c)


def signed_angle_2vec3D(V1:Vec, V2:Vec, N:Vec) -> float:
    """Signed angle between two vectors with orientation given by normal N

    Parameters:
        V1 (Vec): First vector
        V2 (Vec): Second vector
        N (Vec): reference normal direction

    Returns:
        float: the angle
    """
    S = cross(V1,V2)
    s = S.norm()
    c = dot(V1,V2)
    return sign0(dot(S,N)) * math.atan2(s, c)

def signed_angle_3pts(A:Vec, B:Vec, C:Vec, N:Vec) -> float:
    """Signed angle between three points ABC with orientation givne by normal N

    Parameters:
        A (Vec): first point
        B (Vec): central point
        C (Vec): second point
        N (Vec): reference normal direction

    Returns:
        float: the angle
    """
    return signed_angle_2vec3D(A-B, C-B, N)

def angle_2vec2D(V1:Vec, V2:Vec) -> float:
    return math.atan2(V2.y, V2.x) - math.atan2(V1.y, V1.x)

def angle_2vec3D(V1:Vec, V2:Vec) -> float:
    C = cross(V1,V2)
    return math.atan2(C.norm(), dot(V1,V2))

def face_basis(*f):
    """ 
    Orthonormal basis of face
    Given three points A,B,C, returns a basis such that the first vector is along direction AB and third vector is normal to the plane ABC
    """
    if len(f)==1: f = f[0]
    pA,pB,pC = (x for x in f)
    X = Vec.normalized(pB-pA)
    Z = Vec.normalized(cross(X, pC-pA))
    Y = Vec.normalized(cross(Z,X))
    return X,Y,Z

def triangle_area(A:Vec, B:Vec, C:Vec) -> float :
    return cross(B-A,C-A).norm()/2

def triangle_area_2D(A:Vec, B:Vec, C:Vec) -> float :
    return abs(det_2x2(B-A,C-A)/2)

def quad_area(A:Vec, B:Vec, C:Vec, D:Vec) -> float:
    return (triangle_area(A,B,C) + triangle_area(A,C,D) + triangle_area(B,C,D) + triangle_area(B,D,A))/2

def det_2x2(A:Union[complex,np.ndarray], B:Union[complex,np.ndarray]) -> float:
    """Computes a 2x2 determinant

    Parameters:
        A (Union[complex,np.ndarray]): first column vector. Can also be a complex number
        B (Union[complex,np.ndarray]): second column vector. Can also be a complex number

    Returns:
        float: the determinant
    """
    if isinstance(A,complex):
        ax,ay = A.real, A.imag
    else:
        ax,ay = A[0], A[1]
    if isinstance(B,complex):
        bx,by = B.real, B.imag
    else:
        bx,by = B[0], B[1]
    return ax*by - ay*bx

def det_3x3(*args) -> float:
    """Computes a 3x3 using the rule of Sarrus
    
    Parameters:
        Either a 3*3 numpy array representing a matrix, or 3 3*1 numpy array representing three column vectors

    Returns:
        float: the determinant
    """
    if len(args)==1:
        mat = args[0]
        assert mat.shape == (3,3)
    elif len(args)==3:
        A,B,C = args[0], args[1], args[2]
        mat = np.array([A,B,C])
    d = mat[0,0] * mat[1,1] * mat[2,2] + \
        mat[0,1] * mat[1,2] * mat[2,0] + \
        mat[0,2] * mat[1,0] * mat[2,1] - \
        mat[0,0] * mat[1,2] * mat[2,1] - \
        mat[0,1] * mat[1,0] * mat[2,2] - \
        mat[0,2] * mat[1,1] * mat[2,0]
    return d

def intersect_2lines2D(p1 : Vec, d1 : Vec, p2: Vec, d2 : Vec) -> Vec:
    p1,d1,p2,d2 = (u[:2] for u in (p1,d1,p2,d2))
    if abs(det_2x2(d1,d2))<1e-12 : return None #parallel lines
    n2 = Vec(d2.y, -d2.x)
    t = dot(p2-p1,n2)/dot(d1,n2)
    return p1+t*d1

def circumcenter(v1 : Vec, v2 : Vec, v3: Vec) -> Vec:
    """Circumcenter of the triangle formed by three points in space

    /!\ circumcenter of triangle (v1,v2,v3) may not lay inside the triangle

    Parameters:
        v1 (Vec): first point
        v2 (Vec): second point
        v3 (Vec): third point

    Returns:
        Vec: coordinates of the circumcenter
    """
    p1 = (v1+v2)/2
    p2 = (v1+v3)/2
    d1 = v2-v1
    d2 = v3-v1
    d1 = Vec(d1.y, -d1.x)
    d2 = Vec(d2.y, -d2.x)
    S = intersect_2lines2D(p1, d1, p2, d2)
    return Vec(S.x, S.y, 0.)

def aspect_ratio(A : Vec, B : Vec, C : Vec) -> float:
    """
    Computes the aspect ratio of triangle ABC, defined as the ratio between the circumradius to twice the inradius.
    This ratio equals 1 for equilateral triangle and goes to zero as the triangle gets close to degenerate.

    Args:
        A (Vec): first point of the triangle
        B (Vec): second point of the triangle
        C (Vec): third point of the triangle

    Returns:
        float: the aspect ratio of triangle ABC

    Note:
        https://stackoverflow.com/a/10290011
    """
    ab = distance(A,B)
    bc = distance(B,C)
    ca = distance(C,A)
    s = (ab+bc+ca)/2
    return ab*bc*ca/(8*(s-ab)*(s-bc)*(s-ca))


from mouette import framefield as ff
from mouette.processing import SurfaceConnectionFaces, SurfaceConnectionVertices, FeatureEdgeDetector
from mouette.mesh.datatypes.type_checks import BadMeshTypeException
from mouette.utils.argument_check import InvalidArgumentTypeError,InvalidArgumentValueError, InvalidRangeArgumentError
from data import *

@pytest.mark.parametrize("m",[surf_spline(), surf_half_sphere()])
def test_connection_faces(m):
    conn = SurfaceConnectionFaces(m)
    assert True

@pytest.mark.parametrize("m",[surf_spline(), surf_half_sphere()])
def test_connection_vertices(m):
    conn = SurfaceConnectionVertices(m)
    assert True

def test_surface_framefield_wrong_mesh():
    m = M.mesh.new_polyline()
    try:
        _ = ff.SurfaceFrameField(m, "vertices")
        assert False
    except BadMeshTypeException as e:
        assert True

def test_surface_framefield_wrong_elements():
    m = M.mesh.new_surface()
    try:
        _ = ff.SurfaceFrameField(m, 42)
        assert False
    except InvalidArgumentTypeError as e:
        assert True
    
    try:
        _ = ff.SurfaceFrameField(m,"foo")
        assert False
    except InvalidArgumentValueError as e:
        assert True


def test_surface_frame_field_bad_numerical_values():
    m = M.mesh.new_surface()
    try:
        _ = ff.SurfaceFrameField(m, "vertices", n_smooth="pouet")
        assert False
    except InvalidArgumentTypeError as e:
        assert True

    try:
        _ = ff.SurfaceFrameField(m,"faces",n_smooth=-1)
        assert False
    except InvalidRangeArgumentError as e:
        assert True

@pytest.mark.parametrize("m", [surf_circle(), surf_spline(), surf_half_sphere(), surf_pointy()])
def test_surface_framefield_vertices(m):
    framefield = ff.SurfaceFrameField(m, "vertices")()
    framefield.flag_singularities()
    assert m.faces.has_attribute("singuls")

@pytest.mark.parametrize("m", [surf_pointy()])
def test_CADFF(m):
    framefield = ff.SurfaceFrameField(m, "vertices", cad_correction=True)()
    framefield.flag_singularities()
    assert m.faces.has_attribute("singuls")

@pytest.mark.parametrize("m", [surf_circle(), surf_spline(), surf_half_sphere(), surf_pointy()])
def test_surface_framefield_faces(m):
    framefield = ff.SurfaceFrameField(m, "faces")()
    framefield.flag_singularities()
    assert m.vertices.has_attribute("singuls")

@pytest.mark.parametrize("m", [surf_circle()])
def test_surface_framefield_custom_connection_vertices(m):
    conn = SurfaceConnectionVertices(m)
    _ = ff.SurfaceFrameField(m, "vertices", custom_connection=conn)()
    assert True

@pytest.mark.parametrize("m", [surf_circle()])
def test_surface_framefield_custom_connection_faces(m):
    conn = SurfaceConnectionFaces(m)
    _ = ff.SurfaceFrameField(m, "faces", custom_connection=conn)()
    assert True

@pytest.mark.parametrize("m", [surf_feat()])
def test_surface_framefield_custom_features_vertices(m):
    feat = FeatureEdgeDetector()
    _ = ff.SurfaceFrameField(m, "vertices", custom_feature=feat)()
    assert True

@pytest.mark.parametrize("m", [surf_feat()])
def test_surface_framefield_custom_features_faces(m):
    feat = FeatureEdgeDetector()
    _ = ff.SurfaceFrameField(m, "faces", custom_feature=feat)()
    assert True

@pytest.mark.parametrize("m", [surf_circle()])
def test_surface_framefield_custom_singus_vertices(m):
    singus = m.faces.create_attribute("singuls", float)
    singus[0] = 1
    singus[10] = 1
    singus[45] = 1
    singus[76] = 1
    _ = ff.SurfaceFrameField(m, "vertices", singularity_indices=singus)()
    assert True

@pytest.mark.parametrize("m", [surf_circle()])
def test_surface_framefield_custom_singus_faces(m):
    singus = m.vertices.create_attribute("singuls", float)
    singus[64] = 1
    singus[80] = 1
    singus[142] = 1
    singus[150] = 1
    _ = ff.SurfaceFrameField(m, "faces", singularity_indices=singus)()
    assert True

@pytest.mark.parametrize("m", [surf_circle()])
def test_surface_framefield_custom_singus_faces_invalid_topo(m):
    singus = m.vertices.create_attribute("singuls", float)
    singus[64] = 1
    singus[80] = -1 # singus do not verify Poincarré-Hopf
    try:
        _ = ff.SurfaceFrameField(m, "faces", singularity_indices=singus)()
        assert False
    except Exception as e:
        assert True

@pytest.mark.parametrize("m", [surf_circle(), surf_spline(), surf_half_sphere(), surf_pointy()])
def test_CurvatureVertices(m):
    framefield = ff.PrincipalDirections(m, "vertices")()
    framefield.flag_singularities()
    assert m.faces.has_attribute("singuls")

@pytest.mark.parametrize("m", [surf_circle(), surf_spline(), surf_half_sphere(), surf_pointy()])
def test_CurvatureFaces(m):
    framefield = ff.PrincipalDirections(m, "faces")()
    framefield.flag_singularities()
    assert m.vertices.has_attribute("singuls")


import mouette as M
from data import *
import pytest

@pytest.mark.parametrize("m", surfaces)
def test_feature_detect(m):
    feat = M.processing.FeatureEdgeDetector(only_border= False, flag_corners=True)
    feat.run(m)
    assert isinstance(feat.feature_vertices, set)
    assert isinstance(feat.feature_graph, M.mesh.PolyLine)

@pytest.mark.parametrize("m", [surf_spline(), surf_half_sphere()])
def test_feature_detect_only_border(m):
    feat = M.processing.FeatureEdgeDetector(only_border=True, flag_corners=False, compute_feature_graph=False, verbose=False)
    feat.run(m)
    for v in feat.feature_vertices:
        assert m.is_vertex_on_border(v)
    for e in feat.feature_edges:
        assert m.is_edge_on_border(*m.edges[e])


from ..mesh.datatypes import *
from ..mesh.mesh_attributes import ArrayAttribute
from .. import attributes
from .features import FeatureEdgeDetector
from ..geometry import Vec
from .. import geometry as geom
from .. import utils

import numpy as np
import math
from abc import ABC, abstractmethod

class SurfaceConnection(ABC):
    """
    Abstract base class for surface connections.
    """
    @allowed_mesh_types(SurfaceMesh)
    def __init__(self, mesh : SurfaceMesh, feat : FeatureEdgeDetector = None ):
        self.mesh = mesh
        self.feat = feat or FeatureEdgeDetector(only_border=True, compute_feature_graph=False, verbose=False)(self.mesh)
        self._baseX : np.ndarray = None
        self._baseY : np.ndarray = None
        self._transport : dict = None
        self._initialize()
    
    @abstractmethod
    def _initialize(self):
        pass

    def transport(self, iA : int, iB : int) -> float:
        return self._transport[(iA,iB)]

    def base(self, i):
        return self._baseX[i], self._baseY[i]

class SurfaceConnectionVertices(SurfaceConnection):
    """
    Local bases and parallel transport defined between tangent planes of the mesh at the vertices

    Reference:
        Globally Optimal Direction Fields by Knöppel et al. (2013)
    
    Args:
        mesh (SurfaceMesh): the supporting mesh
        feat (FeatureEdgeDetector, optional): feature edges of the mesh. If not provided, will be initialized with a FeatureEdgeDetector object that flags only the boundary. Defaults to None.
    Additionnal args:
        vnormals (Attribute): An attribute representing the vertex normals. If not provided, will be computed at initialization
        angles (Attribute) : An attribute representing angles at corners of faces. If not provided, will be computed at initialization

    """

    @allowed_mesh_types(SurfaceMesh)
    def __init__(self, mesh : SurfaceMesh,  feat : FeatureEdgeDetector = None, **kwargs):
        self.vnormals = kwargs.get("vnormals", None)
        
        if self.vnormals is None: 
            if mesh.vertices.has_attribute("normals"):
                self.vnormals = mesh.vertices.get_attribute("normals")
            else :
                self.vnormals = attributes.vertex_normals(mesh)
        
        self.angles = kwargs.get("angles", None)
        if self.angles is None:
            if mesh.face_corners.has_attribute("angles"):
                self.angles = mesh.face_corners.get_attribute("angles")
            else:
                self.angles = attributes.corner_angles(mesh)
        
        self.total_angle = ArrayAttribute(float, len(mesh.vertices))
        for iC, C in enumerate(mesh.face_corners):
            self.total_angle[C] = self.total_angle[C] + self.angles[iC] 
        super().__init__(mesh, feat)

    def _initialize(self):
        n_vert = len(self.mesh.vertices)
        self._baseX = ArrayAttribute(float, n_vert, 3)
        self._baseY = ArrayAttribute(float, n_vert, 3)
        self._transport = dict()

        for u in range(n_vert):
            # find basis vector -> first edge
            P = Vec(self.mesh.vertices[u])
            N = self.vnormals[u]
            # extract basis edge
            v = self.mesh.connectivity.vertex_to_vertex(u)[0]
            E = self.mesh.vertices[v] - P
            X = Vec.normalized(E - np.dot(E,N)*N) # project on tangent plane
            self._baseX[u] = X
            self._baseY[u] = geom.cross(N,X)

            vert_u = self.mesh.connectivity.vertex_to_vertex(u)
            
            # initialize angles of every edge in this basis
            ang = 0.
            if u in self.feat.feature_vertices:
                dfct = self.feat.corners[u] * 2 * np.pi / self.feat.corner_order # target defect (multiple of pi/order)
                for v in vert_u:
                    T = self.mesh.half_edges.adj(u,v)[0]
                    self._transport[(u,v)] = ang * dfct / self.total_angle[u]
                    c = self.mesh.connectivity.vertex_to_corner_in_face(u,T)
                    if c is None: continue
                    ang += self.angles[c]
            else:
                # normal vertex in interior -> flatten to 2pi
                for v in vert_u :
                    T = self.mesh.half_edges.adj(u,v)[0]
                    self._transport[(u,v)] = ang * 2 * np.pi / self.total_angle[u]
                    c = self.mesh.connectivity.vertex_to_corner_in_face(u,T)
                    ang += self.angles[c]

            # ang = 0.
            # # if u in self.feat.feature_vertices:
            # if self.mesh.is_vertex_on_border(u) :
            #     fst, lst = vert_u[0], vert_u[-1]
            #     pfst, plst = (self.mesh.vertices[x] for x in (fst, lst))
            #     comp_angle = geom.signed_angle_3pts(plst,P,pfst, N) # complementary angle, ie "exterior" angle between two edges on the boundary
            #     comp_angle = 2*np.pi + comp_angle if comp_angle<0 else comp_angle
                
            #     for v in vert_u:
            #         T = self.mesh.half_edges.adj(u,v)[0]
            #         self._transport[(u,v)] = ang * 2 * np.pi / (self.total_angle[u] + comp_angle)
            #         if T is None : continue
            #         c = self.mesh.connectivity.vertex_to_corner_in_face(u,T)
            #         ang += self.angles[c]
            # else:
            #     for v in vert_u:
            #         T = self.mesh.half_edges.adj(u,v)[0]
            #         c = self.mesh.connectivity.vertex_to_corner_in_face(u,T)
            #         self._transport[(u,v)] = ang * 2 * np.pi / self.total_angle[u]
            #         ang += self.angles[c]

class SurfaceConnectionFaces(SurfaceConnection):
    """
    Local bases and parallel transport defined between tangent planes of the mesh at the vertices

    Reference:
        Globally Optimal Direction Fields by Knöppel et al. (2013)

    Args:
        mesh (SurfaceMesh): the supporting mesh
        feat (FeatureEdgeDetector, optional): feature edges of the mesh. If not provided, will be initialized with a FeatureEdgeDetector object that flags only the boundary. Defaults to None.
    """
    @allowed_mesh_types(SurfaceMesh)
    def __init__(self, mesh : SurfaceMesh, feat : FeatureEdgeDetector = None) -> None:
        super().__init__(mesh, feat)

    def _initialize(self):
        #### Initialize bases
        NF = len(self.mesh.faces)
        self._baseX, self._baseY = ArrayAttribute(float, NF, 3), ArrayAttribute(float, NF, 3)
        self._transport = dict()
        for id_face, (A,B,C) in enumerate(self.mesh.faces):
            feat = [self.mesh.connectivity.edge_id(_u,_v) in self.feat.feature_edges for (_u,_v) in [(A,B), (B,C), (C,A)]]
            if np.any(feat):
                # face has feature edge -> feature should be the X coordinate of the basis
                A,B,C = utils.offset([A,B,C], np.argmax(feat))
            pA,pB,pC = (self.mesh.vertices[_v] for _v in (A,B,C))
            X,Y,_ = geom.face_basis(pA,pB,pC) # local basis of the triangle (ignore normal)
            self._baseX[id_face] = X
            self._baseY[id_face] = Y

        #### Initialize Parallel Transport
        for e in self.mesh.interior_edges:
            A,B = self.mesh.edges[e]
            pA,pB = self.mesh.vertices[A], self.mesh.vertices[B]
            E = geom.Vec(pB-pA)
            T1,T2 = self.mesh.half_edges.edge_to_triangles(A,B)
            X1,Y1 = self._baseX[T1], self._baseY[T1]
            X2,Y2 = self._baseX[T2], self._baseY[T2]
            angle1 = math.atan2( geom.dot(E,Y1), geom.dot(E,X1))
            angle2 = math.atan2( geom.dot(E,Y2), geom.dot(E,X2))
            self._transport[(T1,T2)] = angle1 - angle2
            self._transport[(T2,T1)] = angle2 - angle1


import mouette as M
from mouette.processing import framefield
import sys

if len(sys.argv)<2:
    print("Usage: `python framefield3D.py <path/to/mesh>")

m = M.mesh.load(sys.argv[1])
ff = framefield.FrameField3DCells(m, verbose=True)
ff.run() # Initializes values 
ff.flag_singularities() # Computes the singularity graph

M.mesh.save(m, "output.geogram_ascii")
M.mesh.save(ff.export_as_mesh(), "output_FF.geogram_ascii")


import pytest
import mouette as M
import numpy as np
import scipy.sparse as sp

@pytest.fixture
def optim():
    opt = M.optimize.LevenbergMarquardt()
    opt.verbose_options.logger_verbose = False
    return opt

def test_LM1(optim):
    optim.register_function(
        lambda X : (2*X, 2*sp.eye(X.size, dtype=np.float32)),
        lambda X : 2*X,
        0.5,
        "test1")
    X0 = np.full(100, 1.)
    en = optim.run(X0)
    assert en< optim.HP.ENERGY_MIN
    assert np.all( abs(optim.X)<1e-6)

def test_LM_only1fun(optim):
    optim.register_function(
        lambda X : (2*X, 2*sp.eye(X.size, dtype=np.float32)),
        weight=1,
        name="test2")
    X0 = np.full(42, 1.)
    en = optim.run(X0)
    assert en< optim.HP.ENERGY_MIN
    assert np.all( abs(optim.X)<1e-6)

def test_LM_constraints1(optim):
    optim.register_function(
        lambda X : (2*X, 2*sp.eye(X.size, dtype=np.float32)),
        weight=1,
        name="test2")
    
    A = np.zeros((2,42))
    A[0,0]  = 1
    A[1,16] = 1
    A = sp.csc_matrix(A)
    B = np.zeros(2) #x0 = x15 = 0
    optim.register_constraints(A,B,B)
    en = optim.run(x_init=np.full(42, 1.))
    print(en, optim.X)
    assert abs(en)< 1e-4
    assert abs(optim.X[0])<1e-4
    assert abs(optim.X[16])<1e-4

def test_LM_constraints2(optim):
    optim.register_function(
        lambda X : (X, sp.eye(X.size, dtype=np.float32)),
        weight=1,
        name="test2")
    A = np.zeros((3,42))
    A[0,0]  =  1 # x0  = 1
    A[1,16] =  1 # x15 = 3
    A[2,16] =  1 
    A[2,20] =  1 # x15 + x19 = 0
    A = sp.csc_matrix(A)
    B = np.array([1,2,0])
    optim.register_constraints(A,B,B)
    en = optim.run(x_init=np.full(42, 1.))
    assert abs(en-4.5)<1e-4
    assert abs(optim.X[0]-1)<1e-4
    assert abs(optim.X[16]-2)<1e-4
    assert abs(optim.X[20]+2)<1e-4

def test_LM2(optim):
    optim.register_function(
        lambda X : (np.array([np.dot(X,X)-1]), sp.csr_matrix(2*X).reshape((1,X.size))),
        lambda X : np.array([np.dot(X,X)-1]),
        1,
        "test3")
    X0 = np.array([1/x for x in range(1,30)])
    en = optim.run(X0)
    assert en< optim.HP.ENERGY_MIN


def test_LM_twofunctions(optim):
    optim.register_function(
        lambda X : (X, sp.eye(X.size, dtype=np.float32)),
        weight=1,
    )
    optim.register_function(
        lambda X : (2*X, 2*sp.eye(X.size, dtype=np.float32)),
        weight=1.2
    )
    en = optim.run(x_init=np.full(42, 1.))
    assert en < optim.HP.ENERGY_MIN  



import mouette as M
import pytest
import os

### Volumes ###

def vol_tetrahedron():
    m = M.mesh.new_volume()
    m.vertices += [
        M.Vec(0., 0., 0.),
        M.Vec(1., 0., 0.),
        M.Vec(0., 0., 1.),
        M.Vec(0., 1., 0.),
    ]
    m.edges += [(0,1), (1,2), (0,2), (0,3), (1,3), (2,3)]
    m.faces += [[0,1,2], [0,1,3], [1,2,3], [2,0,3]]
    m.cells.append([0,1,2,3])
    return m

def vol_join():
    return M.mesh.load(os.path.join("tests/data/join.tet"))

def vol_cube():
    return M.mesh.load(os.path.join("tests/data/cube86.mesh"))

def vol_cuboid():
    return M.mesh.load(os.path.join("tests/data/p01.mesh"))

volumes = [
    vol_tetrahedron(),
    vol_cuboid(),
    vol_join(),
    vol_cube(),
]


from ..mesh_data import RawMeshData
from .base import Mesh
from .pointcloud import PointCloud
from ... import utils

class PolyLine(Mesh): 
    
    def __init__(self, data : RawMeshData = None):
        Mesh.__init__(self, 1, data)
        self.connectivity = PolyLine._Connectivity(self) 

    def __str__(self):
        out = "mouette.mesh.Polyline object\n"
        out += "| {} vertices\n".format(len(self.vertices))
        out += "| {} edges\n".format(len(self.edges))
        return out

    @property
    def id_vertices(self):
        """
        Shortcut for range(len(self.vertices))
        """
        return range(len(self.vertices))

    @property
    def id_edges(self):
        """
        Shortcut for range(len(self.edges))
        """
        return range(len(self.edges))

    class _Connectivity:
        
        def __init__(self, master):
            self.mesh = master

            self._edge_id : dict = None # (v1, v2) -> edge | None
            self._adjV2V : dict = None # vertex -> vertex

        def clear(self):
            self._edge_id = None
            self._adjV2V = None

        def _compute_vertex_adj(self):
            self._adjV2V = dict([(i,set()) for i in self.mesh.id_vertices ])

            for (A,B) in self.mesh.edges:
                assert A != B # should not be possible
                self._adjV2V[A].add(B)
                self._adjV2V[B].add(A)
                    
            for U in self.mesh.id_vertices:
                self._adjV2V[U] = list(self._adjV2V[U])

        def _compute_edge_adj(self):
            self._edge_id = dict()
            for iE,E in enumerate(self.mesh.edges):
                key = utils.keyify(E)
                self._edge_id[key] = iE

        def edge_id(self, V1, V2):
            if self._edge_id is None:
                self._compute_edge_adj()
            key = utils.keyify(V1,V2)
            return self._edge_id.get(key, None)

        def other_edge_end(self, e, v) -> int:
            A,B = self.mesh.edges[e]
            if v==A: return B
            if v==B: return A
            return None

        ##### Vertex to Vertex #####

        def vertex_to_vertex(self, V : int):
            if self._adjV2V is None : 
                self._compute_vertex_adj()
            return self._adjV2V[V]

        def n_VtoV(self, V : int):
            if self._adjV2V is None:
                self._compute_vertex_adj()
            return len(self._adjV2V[V])

        ##### Vertex to Edges #####

        def n_VtoE(self, V : int):
            return self.n_VtoV(V)

        def vertex_to_edge(self, V : int) :
            return [ self.edge_id(V, u) for u in self.vertex_to_vertex(V)]


import mouette as M
from data.surfaces import *
from utils import *

def test_new():
    m = M.mesh.new_surface()
    assert len(m.vertices)==0
    assert len(m.edges)==0
    assert len(m.faces)==0

# ### Io tests ###

@pytest.mark.parametrize("s", surfaces)
def test_io_obj(s, tmp_path):
    assert build_test_io(s, tmp_path, "obj", 2)

@pytest.mark.parametrize("s", surfaces)
def test_io_geogram_ascii(s, tmp_path):
    assert build_test_io(s, tmp_path, "geogram_ascii", 2)

@pytest.mark.parametrize("s", surfaces)
def test_io_off(s, tmp_path):
    assert build_test_io(s, tmp_path, "off", 2)

@pytest.mark.parametrize("s", surfaces)
def test_io_medit(s, tmp_path):
    assert build_test_io(s, tmp_path, "mesh", 2)

### connectivity Tests ###

@pytest.mark.parametrize("s", [surf_triangle()])
def test_adj(s):
    assert s.connectivity.vertex_to_vertex(0) == [1,2]

@pytest.mark.parametrize("s", surfaces)
def test_adj_V2V(s):
    _ = s.connectivity.vertex_to_vertex(0)
    _ = s.connectivity.n_VtoV(0)
    assert True

@pytest.mark.parametrize("s", surfaces)
def test_adj_V2F(s):
    _ = s.connectivity.vertex_to_face(0)
    _ = s.connectivity.n_VtoF(0)
    assert True

@pytest.mark.parametrize("s", surfaces)
def test_adj_F2F(s):
    _ = s.connectivity.face_to_face(0)
    assert True

@pytest.mark.parametrize("s", surfaces)
def test_adj_F2V(s):
    _ = s.connectivity.face_to_vertex(0)
    assert True

@pytest.mark.parametrize("s", surfaces)
def test_adj_F2E(s):
    _ = s.connectivity.face_to_edge(0)
    assert True

@pytest.mark.parametrize("s", surfaces)
def test_adj_F2V(s):
    _ = s.connectivity.face_to_vertex(0)
    assert True

### Boundary / Interior Tests ###

@pytest.mark.parametrize("s, expected", [(surf_square(), 4), (surf_cube(), 0)])
def test_bnd(s, expected):
    assert len(s.boundary_vertices) == expected

@pytest.mark.parametrize("s, expected", [(surf_square(), 0), (surf_cube(), 8)])
def test_interior(s, expected):
    assert len(s.interior_vertices) == expected
